"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const got_1 = tslib_1.__importDefault(require("got"));
const request_1 = require("./request");
const response_1 = require("./response");
const logger_1 = tslib_1.__importDefault(require("../logger"));
/**
 * There is no valid reason to have more than one driver instances per express app
 */
const appSet = new Set();
const asyncNoop = async () => { };
class ExpressDriver {
    constructor({ app, env }) {
        logger_1.default.debug(`instantiating new driver`);
        if (appSet.has(app)) {
            throw new Error(`express app must be used as singleton`);
        }
        this.app = app;
        this.env = env;
        this.isActive = true;
        // `/tinkoffApi/nearest_region?...` → `${env.tinkoffApi}/nearest_region?...`
        const resolveReal = (originalUrl, api, key) => {
            const url = api + originalUrl.replace(new RegExp(`/${key}`), '');
            // It could happen `//` – replace it with `/`
            return url.replace(/([^:])[/]+/g, '$1/');
        };
        for (let key in this.env) {
            const apiUrl = this.env[key];
            logger_1.default.info(`add proxy "/${key}" → "${apiUrl}"`);
            this.app.use('/' + key, async (req, res) => {
                var _a, _b, _c, _d;
                logger_1.default.debug(`enter /${key} router with url "${req.originalUrl}"`, key);
                const realUrl = resolveReal(req.originalUrl, apiUrl, key);
                const getRealResponse = async (r) => {
                    var _a, _b;
                    // omit requester-specific data, since it can break request
                    const { host, origin, referer, ...rest } = r.headers || {};
                    const opts = {
                        method: r.method,
                        headers: rest,
                        throwHttpErrors: false,
                    };
                    if (r.body && r.method.toLowerCase() !== 'get') {
                        switch ((_a = r.headers) === null || _a === void 0 ? void 0 : _a['content-type']) {
                            case 'application/x-www-form-urlencoded':
                                // we already converted r.body from string to obj in the express body parser middleware
                                // so, we need to convert it back to string
                                opts.body = new URLSearchParams(r.body).toString();
                                break;
                            default:
                                opts.body = typeof r.body === 'string' ? r.body : JSON.stringify(r.body);
                        }
                        if ((_b = r.headers) === null || _b === void 0 ? void 0 : _b['content-length']) {
                            opts.headers['content-length'] = Buffer.byteLength(opts.body, 'utf-8');
                        }
                    }
                    try {
                        const gotResponse = await got_1.default(realUrl, opts);
                        return response_1.extractGotResponse(gotResponse);
                    }
                    catch (e) {
                        return {
                            url: 'error',
                            status: 500,
                            body: `got error message: ${e.message}`,
                        };
                    }
                };
                const driverReq = await request_1.extractExpressRequest(req, res, realUrl, getRealResponse);
                const { request, onRespondPromise } = driverReq;
                // const timestamp = Date.now()
                logger_1.default.debug(`realUrl "${realUrl}"`);
                if (this.isActive) {
                    (_a = this.onRequestHandler) === null || _a === void 0 ? void 0 : _a.call(this, driverReq);
                    try {
                        const resolved = await onRespondPromise;
                        const interceptor = resolved.interceptor;
                        let sendResp;
                        if ('response' in resolved) {
                            const { response } = resolved;
                            sendResp = response;
                            (_b = this.onResponseHandler) === null || _b === void 0 ? void 0 : _b.call(this, {
                                request,
                                response: {
                                    ...response,
                                },
                                __meta: {
                                    request,
                                    interceptor,
                                },
                            });
                        }
                        else {
                            const { realResponse } = resolved;
                            sendResp = realResponse;
                            // const t2 = Date.now()
                            logger_1.default.debug(`realResponse`, realResponse);
                            (_c = this.onResponseHandler) === null || _c === void 0 ? void 0 : _c.call(this, {
                                request,
                                response: {
                                    ...realResponse,
                                    // ttfb: t2 - timestamp,
                                },
                                __meta: {
                                    request,
                                    interceptor,
                                },
                            });
                        }
                        // it doesn't matter what content-encoding was originally,
                        // teremock has already decoded the response
                        sendResp.headers = (_d = sendResp.headers) !== null && _d !== void 0 ? _d : {};
                        sendResp.headers['content-encoding'] = 'identity';
                        delete sendResp.headers['x-content-encoding-over-network'];
                        res.status(sendResp.status).set(sendResp.headers).send(sendResp.body);
                    }
                    catch (e) {
                        logger_1.default.error(e.message);
                    }
                }
                else {
                    res.status(500).send('driver not active');
                }
            });
        }
    }
    setRequestInterception(arg) {
        this.isActive = arg;
    }
    async onRequest(fn) {
        this.onRequestHandler = fn;
        return async () => {
            this.setRequestInterception(false);
            appSet.delete(this.app);
            this.onRequestHandler = asyncNoop;
        };
    }
    async onResponse(fn) {
        this.onResponseHandler = fn;
        return () => {
            this.onResponseHandler = () => Promise.resolve();
        };
    }
    onClose(fn) {
        return fn;
    }
}
exports.default = ExpressDriver;
