"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const sprout_data_1 = require("sprout-data");
// @ts-ignore
const url_1 = require("url");
// @ts-ignore
const query_string_1 = tslib_1.__importDefault(require("query-string"));
const words_hash_1 = require("./words-hash");
// Length of query string after which three-words-naming is switching on
const MAX_QUERY_NAME_LENGTH = 25;
function pickDeep(obj, keys) {
    return keys.reduce((acc, key) => {
        let val;
        if (Array.isArray(key)) {
            val = sprout_data_1.get(obj, key);
            if (typeof val !== 'undefined') {
                sprout_data_1.assoc(acc, key, val);
            }
        }
        else {
            val = sprout_data_1.get(obj, [key]);
            sprout_data_1.assoc(acc, [key], val);
        }
        return acc;
    }, {});
}
/**
 * Request Id generator.
 * Accounts for request data:
 * 1. url (protocol + host + path)
 * 2. query params (filtered with naming.query)
 * 3. method (get, post...) (as prefix)
 * 4. body, if any (filtered with naming.body)
 *
 * getRequestId does not accounts for
 * 1. blacklisted query and body params
 * 2. Any response data, including status, body and headers
 * 3. request headers (do you need that?)
 * 4. cookies
 *
 * Order agnostic, so, `/foo=bar&baz=1` will have the same rid as `/baz=1&foo=bar`
 */
const getRequestId = (params) => {
    var _a, _b, _c, _d;
    let url = params.url;
    let method = (params.method || 'GET').toLowerCase();
    let headers = params.headers;
    let body = params.body || '';
    let queryWhitelist = ((_a = params.naming.query) === null || _a === void 0 ? void 0 : _a.whitelist) || [];
    let queryBlacklist = ((_b = params.naming.query) === null || _b === void 0 ? void 0 : _b.blacklist) || [];
    let bodyWhitelist = ((_c = params.naming.body) === null || _c === void 0 ? void 0 : _c.whitelist) || [];
    let bodyBlacklist = ((_d = params.naming.body) === null || _d === void 0 ? void 0 : _d.blacklist) || [];
    const urlObj = new url_1.URL(url);
    let bodyObj;
    urlObj.searchParams.sort();
    // @todo remove it from here, rename body to body
    if (body !== '' && headers) {
        if (headers['content-type'] && headers['content-type'].includes('application/json')) {
            bodyObj = JSON.parse(body);
        }
        else {
            bodyObj = query_string_1.default.parse(body);
        }
    }
    if (bodyObj) {
        if (bodyWhitelist.length) {
            bodyObj = pickDeep(bodyObj, bodyWhitelist);
        }
        bodyBlacklist.forEach((param) => {
            let currentObj = bodyObj;
            let paramForDelete = param;
            if (Array.isArray(paramForDelete)) {
                const path = [...param];
                while (path.length > 1) {
                    // @ts-ignore
                    currentObj = currentObj[path.shift()];
                    if (!currentObj) {
                        return;
                    }
                }
                paramForDelete = path.shift();
            }
            delete currentObj[paramForDelete];
        });
        body = JSON.stringify(bodyObj);
    }
    /**
     * If naming.query.whitelist is set, removes any searchParams from the url,
     * which are not declared in the whitelist array
     */
    if (queryWhitelist.length > 0) {
        // @ts-ignore
        let keys = [...urlObj.searchParams.keys()];
        for (let key of keys) {
            if (!queryWhitelist.includes(key)) {
                urlObj.searchParams.delete(key);
            }
        }
    }
    // Some parameters could vary over the time, so we can exclude them from naming
    // (be carefull, use it only if it does not affect actual response body)
    // @ts-ignore
    queryBlacklist.forEach((param) => urlObj.searchParams.delete(param));
    let baseStr = urlObj.toString() + body;
    const entries = [...urlObj.searchParams];
    const queryStr = entries.reduce((acc, [key, value]) => {
        acc.push(`${key}-${value}`);
        return acc;
    }, []).join('-');
    let queryStrAscii = queryStr.toLowerCase().replace(/[^a-z0-9-]/g, '');
    if (method === 'get' && !queryStr.length) {
        return method;
    }
    if (method === 'get' && queryStr.length < MAX_QUERY_NAME_LENGTH && queryStr.length === queryStrAscii.length) {
        return `${method}-${queryStrAscii}`;
    }
    return `${method}-${words_hash_1.humanize(baseStr, 3)}`;
};
exports.default = getRequestId;
